﻿//-----------------------------------------------------------------------
// <copyright file="Registry.cs" company="Quill18 Productions">
//     Copyright (c) Quill18 Productions. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System;
using UnityEngine.SceneManagement;

public class Registry : MonoBehaviour 
{
    static Registry _instance;
    static public Registry current
    {
        get
        {

            return _instance;
        }
    }

    void Awake () 
    {
        _instance = this;

        registry = new Dictionary<string, RegistryItem>();


        //SetValue("MEAT-M", (int)LightMode.BLINK);

        SetupModeListeners();

        BallInScoop = new Dictionary<string, GameObject>();

        modeExpirers = new List<Action>();
    }

    // Scoop.Oven
    // Scoop.Cutting_Board
    // Scoop.Sink
    public Dictionary<string, GameObject> BallInScoop;

    class RegistryItem
    {
        int _value;
        public int Value {
            get { return _value; }
            set {
                _value = value;
                if(listeners != null)
                {
                    listeners();
                }
            }
        }

        public event Action listeners;
    }

    Dictionary<string, RegistryItem> registry;

    public event Action<float> OnUpdate;

    void Update()
    {
        if(Input.GetKeyDown(KeyCode.R))
        {
            Reset();
        }



        if(IsGameOver)
        {
            return;
        }





        if(OnUpdate != null)
        {
            OnUpdate(Time.deltaTime);
        }

        /*if(Input.GetKeyDown(KeyCode.S))
        {
            foreach(string m in MODES)
            {
                SetValue(m + ".INDICATOR", 1);
            }
            SetValue("MODE.WIZARD.INDICATOR", 1);
        }*/

    }

    public bool IsGameOver = false;

    public void Reset()
    {
        SceneManager.LoadScene( SceneManager.GetActiveScene().buildIndex );

    }

    public void GameOver()
    {
        //throw new Exception("GAME OVER");
        IsGameOver = true;

        int highScore = PlayerPrefs.GetInt("High Score");

        if( GetValue("Points") > highScore )
        {
            PlayerPrefs.SetInt("High Score", GetValue("Points"));
        }

        // Delete balls
        GameObject[] balls = GameObject.FindGameObjectsWithTag("Player");
        foreach(GameObject ball in balls)
        {
            Destroy(ball);
        }
    }

    public int GetValue(string name, int def = 0)
    {

        if(registry.ContainsKey(name))
        {
            return registry[name].Value;
        }

        return def;
    }

    public void SetValue(string name, int value)
    {
        if(registry.ContainsKey(name) == false)
        {
            registry[name] = new RegistryItem();
        }

        registry[name].Value = value;
    }

    public void ChangeValue(string name, int amount, int def = 0)
    {
        if(registry.ContainsKey(name) == false)
        {
            registry[name] = new RegistryItem();
            registry[name].Value = def;
        }


        registry[name].Value = GetValue(name, def) + amount;
    }

    public void AddPoints( int amount )
    {
        // TODO: Score multipliers?

        amount *= GetValue("GlobalPointMult", 1);

        SoundManager._instance.ScoreChime(amount);

        ChangeValue("Points", amount);
    }

    string[] MEATPIE = {
        "MEAT-M",
        "MEAT-E",
        "MEAT-A",
        "MEAT-T",
        "PIE-P",
        "PIE-I",
        "PIE-E"
    };

    string[] MODES = {
        "MODE.MEATPIE",
        "MODE.MEATBALL_MULTIBALL",
        "MODE.SUPER_SOUP",
        "MODE.SALT"
    };


    void SetupModeListeners()
    {
        //listeners

        AddListener("RAMP.LEFT.COUNT", OnRampCountChanged);
        AddListener("RAMP.LEFT.COUNT", OnLeftRampChanged);

        AddListener("RAMP.RIGHT.COUNT", OnRampCountChanged);

        AddListener("SCOOP.SINK", OnScoopSinkChanged);

        AddListener("LOOP.RIGHT.COUNT", OnRightLoopChanged);

        AddListener("SCOOP.OVEN", OnWizardProgress_OverScoop);
        AddListener("SCOOP.OVEN", OnMeatballProgress_OvenScoop);

        AddListener("RAMP.RIGHT.COUNT", OnMeatballProgress_RightRamp);
        AddListener("SCOOP.CUTTING_BOARD", OnMeatballProgress_CuttingBoard);

        foreach(string m in MEATPIE)
        {
            AddListener(m, OnMeatpieTrigger);
        }

        foreach(string m in MODES)
        {
            AddListener(m + ".INDICATOR", OnModeChanged);
        }
    }

    void AddListener(string name, Action action)
    {
        if(registry.ContainsKey(name) == false)
        {
            registry[name] = new RegistryItem();
        }

        registry[name].listeners += action;
    }

    void RemoveListener(string name, Action action)
    {
        if(registry.ContainsKey(name) == false)
        {
            return;
        }

        registry[name].listeners -= action;
    }

    Action SetupTimer(float seconds, Action<float> updateAction, Action expireAction)
    {

        float timer = 0;

        Action expireWrapper = null;

        Action<float> updateWrapper = null;
        updateWrapper = (dt) => {
            timer += dt;

            if(updateAction != null)
            {
                updateAction(dt);
            }
            
            if(timer >= seconds)
            {
                expireWrapper();
            }
        };

        OnUpdate += updateWrapper;

        AddListener("BallNum", expireAction);

        expireWrapper = () => {
            OnUpdate -= updateWrapper;
            RemoveListener("BallNum", expireAction);


            if(expireAction != null)
            {
                expireAction();
            }
        };

        return expireWrapper;
    }

    void LockBall( string ScoopName )
    {
        SoundManager._instance.Play( SoundManager._instance.locked );

        if(BallInScoop.ContainsKey(ScoopName) == false)
        {
            Debug.LogError("No scoop named: " + ScoopName);
            return;
        }

        if(BallInScoop[ScoopName] == null)
        {
            Debug.LogError("Scoop is empty: " + ScoopName);
            return;
        }

        Destroy( BallInScoop[ScoopName] );
        ChangeValue("BallsInPlay", -1);
    }

    ///////////////

    List<Action> modeExpirers;

    void StartMode_SuperSoup()
    {
        if( GetValue("MODE.SUPER_SOUP") > 0 )
        {
            return;
        }

        if( GetValue("MODE.WIZARD.SILENCER") == 0 )
            SoundManager._instance.Play( SoundManager._instance.souper_pops );


        Debug.Log("Starting SUPER SOUP mode");
        SetValue("MODE.SUPER_SOUP", 1);
        SetValue("MODE.SUPER_SOUP.INDICATOR", 2);
        SetValue("RAMP.LEFT.COUNT", 0);
        SetValue("RAMP.RIGHT.COUNT", 0);
        ChangeValue("rampCountTarget", 1);

        ChangeValue("POP.POINT_MULTIPLIER", 19, 1 );

        //POP.NUM_HITS
        int initPophits = GetValue("POP.NUM_HITS");
        Action popJackpot = () => {
            if(GetValue("POP.NUM_HITS") >= initPophits+10)
            {
                // Score jackpot!
                Debug.Log("SUPER SOUP:  Jackpot!");
                AddPoints( 1000000 );
                initPophits += 10;
            }
        };
        AddListener("POP.NUM_HITS", popJackpot);

        Action expireAction = () => {
            SetValue("MODE.SUPER_SOUP", 0);
            SetValue("MODE.SUPER_SOUP.INDICATOR", 1);
            ChangeValue("POP.POINT_MULTIPLIER", -19 );
            RemoveListener("POP.NUM_HITS", popJackpot);
        };

        SetupTimer(GetValue("MODE.SUPER_SOUP.TIME_LIMIT", 60), null, expireAction);

    }

    void StartMode_SinkQuickball()
    {
        if( GetValue("MODE.SINK_QUICKBALL") > 0 )
        {
            return;
        }

        Debug.Log("Starting SINK QUICKBALL mode");

        SetValue("MODE.SINK_QUICKBALL", 1);

        SoundManager._instance.Play( SoundManager._instance.multiball    );


        //GameObject.FindObjectOfType<BallSpawner>().SpawnBall(true);

        SetupTimer( 2.5f, null, () => {
            GameObject.Find("SCOOP.SINK").GetComponent<Kickout>().SpawnBall();
        });


        ChangeValue("GlobalPointMult", 1);

        Action ballListener = null;
        ballListener = () => 
            {
                if( GetValue("BallsInPlay") <= 1 )
                {
                    ChangeValue("GlobalPointMult", -1);
                    SetValue("MODE.SINK_QUICKBALL", 0);
                    SetValue("SCOOP.SINK", 0);
                    ChangeValue("MODE.SINK_QUICKBALL_TARGET", 1, 3);
                    RemoveListener("BallsInPlay", ballListener);
                }
            };

        AddListener("BallsInPlay", ballListener);
    }



    void StartMode_MeatballMultiball()
    {
        if( GetValue("MODE.MEATBALL_MULTIBALL") > 0 )
        {
            return;
        }

        Debug.Log("StartMode_MeatballMultiball");

        if( GetValue("MODE.WIZARD.SILENCER") == 0 )
            SoundManager._instance.Play( SoundManager._instance.meatball_multiball );

        SetValue("MODE.MEATBALL_MULTIBALL", 1);

        SetValue("MEATBALL_PROGRESS", 0);
        SetValue("LOCK", 0);

        //GameObject.FindObjectOfType<BallSpawner>().SpawnBall(true);

        // Kick out two extra balls (from the lock)
        ChangeValue("BallsInPlay", 2);
        SetupTimer( 2.5f, null, () => {
            GameObject.Find("SCOOP.OVEN").GetComponent<Kickout>().SpawnBall(false);
        });
        SetupTimer( 4f, null, () => {
            GameObject.Find("SCOOP.OVEN").GetComponent<Kickout>().SpawnBall(false);
        });

        ChangeValue("MULT.RAMP.LEFT", 19, 1);
        ChangeValue("MULT.RAMP.RIGHT", 19, 1);
        ChangeValue("Arrow-Left-Ramp", 2);
        ChangeValue("Arrow-Right-Ramp", 2);

        SetValue("MODE.MEATBALL_MULTIBALL.INDICATOR", 2);

        Action ballListener = null;
        ballListener = () => 
            {
                if( GetValue("BallsInPlay") <= 1 )
                {
                    SetValue("MODE.MEATBALL_MULTIBALL", 0);
                    SetValue("MODE.MEATBALL_MULTIBALL.INDICATOR", 1);
                    ChangeValue("MULT.RAMP.LEFT", -19, 20);
                    ChangeValue("MULT.RAMP.RIGHT", -19, 20);
                    SetValue("SCOOP.OVEN", 0);
                    RemoveListener("BallsInPlay", ballListener);
                    ChangeValue("Arrow-Left-Ramp", 0);
                    ChangeValue("Arrow-Right-Ramp", 0);
                }
            };

        AddListener("BallsInPlay", ballListener);

        //SetupTimer( -1, null, () => { ChangeValue("GlobalPointMult", -1); } );


    }



    void StartMode_Salt()
    {
        if( GetValue("MODE.SALT") > 0 )
        {
            return;
        }

        Debug.Log("StartMode_Salt");
        SetValue("MODE.SALT", 1);

        if( GetValue("MODE.WIZARD.SILENCER") == 0 )
            SoundManager._instance.Play( SoundManager._instance.salty );

        if(saltRampTimerExpirer != null)
        {
            saltRampTimerExpirer();
        }

        ChangeValue("SALT.MULT", 249, 1);
        SetValue("MODE.SALT.INDICATOR", 2);

        Action expireAction = () => 
            {
                SetValue("MODE.SALT", 0);
                ChangeValue("SALT.MULT", -249, 250);
                SetValue("MODE.SALT.INDICATOR", 1);
            };

        SetupTimer(GetValue("MODE.SALT.TIME_LIMIT", 30), null, expireAction);

    }


    void StartMode_Meatpie()
    {
        if( GetValue("MODE.MEATPIE") > 0 )
        {
            return;
        }
        // - Hit Oven for mini-jackpot as many times as you can in 30 seconds

        if( GetValue("MODE.WIZARD.SILENCER") == 0 )
            SoundManager._instance.Play( SoundManager._instance.meatpie_ready );

        SetValue("MODE.MEATPIE", 1);
        SetValue("MODE.MEATPIE.INDICATOR", 2);

        SetValue("Arrow-Oven", 2);
        SetValue("Arrow-Sink", 2);
        SetValue("Arrow-Cutting-Board", 2);

        ChangeValue("MULT.SCOOP.OVEN", 24, 1);
        ChangeValue("MULT.SCOOP.SINK", 24, 1);
        ChangeValue("MULT.SCOOP.CUTTING_BOARD", 24, 1);

        // TODO: Make the shots one-time, but resetable and worth a million each

        Action expireAction = () => 
            {
                SetValue("MODE.MEATPIE", 0);
                SetValue("MODE.MEATPIE.INDICATOR", 1);
                SetValue("Arrow-Oven", 0);
                SetValue("Arrow-Sink", 0);
                SetValue("Arrow-Cutting-Board", 0);

                ChangeValue("MULT.SCOOP.OVEN", -24, 25);
                ChangeValue("MULT.SCOOP.SINK", -24, 25);
                ChangeValue("MULT.SCOOP.CUTTING_BOARD", -24, 25);


                foreach(string m in MEATPIE)
                {
                    SetValue(m, 0);
                }


            };


        SetupTimer(GetValue("MODE.MEATPIE.TIME_LIMIT", 30), null, expireAction);

    }

    void StartMode_Wizard()
    {
        Debug.Log("StartMode_Wizard");
        if( GetValue("MODE.WIZARD") > 0 )
        {
            return;
        }

        SoundManager._instance.Play( SoundManager._instance.wizard_mode );

        SetValue("MODE.WIZARD", 1);

        SetValue("MODE.WIZARD.SILENCER", 1);           
        StartMode_MeatballMultiball();
        StartMode_Salt();
        StartMode_SuperSoup();
        StartMode_Meatpie();
        SetValue("MODE.WIZARD.SILENCER", 0);

        SetValue("MODE.WIZARD.INDICATOR", 2);

        ChangeValue("GlobalPointMult", 9, 1);

        ChangeValue("BallsInPlay", 1);
        SetupTimer( 7.5f, null, () => {
            GameObject.Find("SCOOP.OVEN").GetComponent<Kickout>().SpawnBall(false);
        });

        Action ballListener = null;
        ballListener = () => 
            {
                if( GetValue("BallsInPlay") <= 1 )
                {
                    SetValue("MODE.WIZARD", 0);
                    ChangeValue("GlobalPointMult", -9, 10);
                    SetValue("MODE.WIZARD.INDICATOR", 0);

                    foreach(string m in MODES)
                    {
                        SetValue(m + ".INDICATOR", 0);
                    }
                }
            };

        AddListener("BallsInPlay", ballListener);


    }



    ///////////////

    void OnRampCountChanged()
    {
        Debug.Log("OnRampCountChanged");

        int rampCountTarget = GetValue("rampCountTarget", 3);

        if(
            GetValue("RAMP.LEFT.COUNT") >= rampCountTarget &&
            GetValue("RAMP.RIGHT.COUNT") >= rampCountTarget &&
            GetValue("MODE.SUPER_SOUP") == 0
        )
        {
            StartMode_SuperSoup();
        }
    }

    void OnScoopSinkChanged()
    {
        Debug.Log("OnScoopSinkChanged");

        if( GetValue("SCOOP.SINK") >= GetValue("MODE.SINK_QUICKBALL_TARGET", 3) )
        {
            SetValue("CLOGGED", 0);
            StartMode_SinkQuickball();
        }
        else if( GetValue("SCOOP.SINK") == GetValue("MODE.SINK_QUICKBALL_TARGET", 3)-1 )
        {
            LockBall("SCOOP.SINK");
            SetValue("CLOGGED", 2);
            SetupTimer( 1f, null, () => {
                GameObject.FindObjectOfType<BallSpawner>().SpawnBall(true);
            });
        }
    }

    Action expireMB_Ramp;
    Action expireMB_CuttingBoard;

    void OnMeatballProgress_RightRamp()
    {
        if( GetValue("MODE.MEATBALL_MULTIBALL") > 0 )
        {
            return;
        }


        if(expireMB_CuttingBoard != null)
            expireMB_CuttingBoard();
        
        SetValue("Arrow-Cutting-Board", 2);
        expireMB_CuttingBoard = SetupTimer(30, null, () => { SetValue("Arrow-Cutting-Board", 0); } );

        if( GetValue("Arrow-Right-Ramp") > 0 ) {
            SetValue("Arrow-Right-Ramp", 0);
            ChangeValue("MEATBALL_PROGRESS", 1);

            if( GetValue("MEATBALL_PROGRESS") >= 3 )
            {
                SetValue("LOCK", 2);
            }
        }
    }

    void OnMeatballProgress_CuttingBoard()
    {
        if( GetValue("MODE.MEATBALL_MULTIBALL") > 0 )
        {
            return;
        }


        if(expireMB_Ramp != null)
            expireMB_Ramp();

        SetValue("Arrow-Right-Ramp", 2);
        expireMB_Ramp = SetupTimer(30, null, () => { SetValue("Arrow-Right-Ramp", 0); } );

        if( GetValue("Arrow-Cutting-Board") > 0 ) {
            SetValue("Arrow-Cutting-Board", 0);
            ChangeValue("MEATBALL_PROGRESS", 1);

            if( GetValue("MEATBALL_PROGRESS") >= 3 )
            {
                SetValue("LOCK", 2);
            }
        }
    }

/*    void OnMeatballProgress()
    {
        if( GetValue("MODE.MEATBALL_MULTIBALL") > 0 )
        {
            return;
        }

        Debug.Log("OnMeatballProgress");

        if( GetValue("SCOOP.CUTTING_BOARD") > GetValue("SCOOP.CUTTING_BOARD.MEATBALL") && 
            GetValue("RAMP.RIGHT") > GetValue("RAMP.RIGHT.MEATBALL") )
        {
            ChangeValue("MEATBALL_PROGRESS", 1);
            SetValue("SCOOP.CUTTING_BOARD.MEATBALL", GetValue("SCOOP.CUTTING_BOARD"));
            SetValue("RAMP.RIGHT.MEATBALL", GetValue("RAMP.RIGHT"));

            if( GetValue("MEATBALL_PROGRESS") >= 3 )
            {
                SetValue("LOCK", 2);
            }
        }
    }
*/
    void OnMeatballProgress_OvenScoop()
    {
        //Debug.Log("OnMeatballProgress_OvenScoop");
        if(GetValue("LOCK") > 0)
        {
            // locking balls for Meatball Multiball

            ChangeValue("MEATBALL.LOCKED", 1);

            if(GetValue("MEATBALL.LOCKED") >= 3)
            {
                StartMode_MeatballMultiball();
            }
            else
            {
                LockBall("SCOOP.OVEN");
                SetupTimer( 1f, null, () => {
                    GameObject.FindObjectOfType<BallSpawner>().SpawnBall(true);
                });
            }
        }
    }

    Action saltRampTimerExpirer;

    void OnRightLoopChanged()
    {
        Debug.Log("OnRightLoopChanged");

        if( GetValue("MODE.SALT") > 0)
        {
            return;
        }

        if( GetValue("MODE.SALT.LEFT_RAMP_ACTIVE") > 0 )
        {
            saltRampTimerExpirer();
        }

        SetValue("MODE.SALT.LEFT_RAMP_ACTIVE", 1);

        SetValue("Arrow-Left-Ramp", 2);

        Action expireAction = null;

        expireAction = () => {
            SetValue("MODE.SALT.LEFT_RAMP_ACTIVE", 0);
            SetValue("Arrow-Left-Ramp", 0);
        };

        saltRampTimerExpirer = SetupTimer(30, null, expireAction);
    }

    void OnLeftRampChanged()
    {
        Debug.Log("OnLeftRampChanged");
        if( GetValue("MODE.SALT.LEFT_RAMP_ACTIVE") > 0 )
        {
            StartMode_Salt();
        }
    }

    void OnMeatpieTrigger()
    {


        foreach(string m in MEATPIE)
        {
            if( GetValue(m) <= 0 )
            {
                return;
            }
        }

        // MEATPIE is all lit!

        // - Hit Oven for mini-jackpot as many times as you can in 30 seconds
        StartMode_Meatpie();
    }


    void OnModeChanged()
    {
        //Debug.Log("OnModeChanged");
        // Checking for WIZARD MODE readiness
        if( GetValue("MODE.WIZARD") > 0 )
        {
            return;
        }

        // Check that all the indicators equal 1 (completed, but not running)
        foreach(string s in MODES)
        {
            string s2 = s + ".INDICATOR";
            if( GetValue(s2) != 1 )
            {
                Debug.Log(s2 + " isn't ready for wizard mode. " + GetValue(s2));
                return;
            }
        }

        // Wizard mode ready!
        SetValue("MODE.WIZARD.INDICATOR", 1);
        SetValue("Arrow-Oven", 2);
    }

    void OnWizardProgress_OverScoop()
    {
        //Debug.Log("OnWizardProgress_OverScoop");
        if( GetValue("MODE.WIZARD") > 0 || GetValue("MODE.WIZARD.INDICATOR") == 0 )
        {
            return;
        }

        StartMode_Wizard();

    }
}
