﻿//-----------------------------------------------------------------------
// <copyright file="Indicator.cs" company="Quill18 Productions">
//     Copyright (c) Quill18 Productions. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public enum LightMode { OFF, SOLID, BLINK };

public class Indicator : MonoBehaviour 
{

    void Start () 
    {

        ResetTexture();
    }

    public Material IndicatorMaterial;

    public Texture Texture;

    public LightMode DebugLightMode;

    bool isLit;
    float toggleIn = 0;
    float blinkDelay = 0.15f;

    MeshRenderer meshRenderer;

    [ContextMenu ("Reset Texture")]
    public void ResetTexture()
    {
        if(meshRenderer == null)
        {
            meshRenderer = GetComponentInChildren<MeshRenderer>();
        }

        Material mat = new Material(IndicatorMaterial);
        mat.mainTexture = Texture;
        mat.SetTexture("_EmissionMap", Texture);
        meshRenderer.material = mat;
        TurnOff();
    }

    void TurnOff()
    {
        isLit = false;
        Material mat = meshRenderer.material;
        mat.SetColor("_EmissionColor", Color.black);
        //meshRenderer.material = mat;
    }

    void TurnOn()
    {
        isLit = true;
        Material mat = meshRenderer.material;
        mat.SetColor("_EmissionColor", Color.white);
        //meshRenderer.material = mat;
    }

    void Update()
    {
        LightMode lm = (LightMode)Registry.current.GetValue(gameObject.name);

        if(DebugLightMode != LightMode.OFF)
        {
            lm = DebugLightMode;
        }

        if(lm == LightMode.OFF && isLit == true)
        {
            TurnOff();
        }
        else if(lm == LightMode.SOLID && isLit == false)
        {
            TurnOn();
        }
        else if(lm == LightMode.BLINK)
        {
            toggleIn -= Time.deltaTime;
            if(toggleIn <= 0)
            {
                if(isLit)
                    TurnOff();
                else
                    TurnOn();

                toggleIn = blinkDelay;
            }
        }
    }
    
}
